@@ -24,6 +24,7 @@
  * Google Author(s): Garret Rieger
  */
 
+#include <cstdint>
 #include <string>
 
 #include "hb-repacker.hh"
@@ -147,6 +148,17 @@ static unsigned add_extension (unsigned child,
 static unsigned add_coverage (unsigned start, unsigned end,
                               hb_serialize_context_t* c)
 {
+  if (end - start == 0) {
+    uint8_t coverage[] = {
+      0, 1, // format
+      0, 1, // count
+
+      (uint8_t) ((start >> 8) & 0xFF),
+      (uint8_t) (start & 0xFF), // glyph[0]
+    };
+    return add_object ((char*) coverage, 6, c);
+  }
+
   if (end - start == 1)
   {
     uint8_t coverage[] = {
@@ -242,6 +254,31 @@ static unsigned add_pair_pos_1 (unsigned* pair_sets,
   return c->pop_pack (false);
 }
 
+static void add_liga_set_header (char liga_count,
+                                 hb_serialize_context_t* c)
+{
+  uint8_t liga_count_bytes[] = {
+    (uint8_t) (0xFF & (liga_count >> 8)),
+    (uint8_t) (0xFF & liga_count),
+  };
+
+  start_object ((const char*) liga_count_bytes, 2, c);
+}
+
+static void add_liga_header (unsigned coverage, unsigned liga_set_count, hb_serialize_context_t* c)
+{
+  uint8_t format[] = {0, 1};
+  start_object ((const char*) format, 2, c);
+  add_offset(coverage, c);
+
+  uint8_t liga_set_count_bytes[] = {
+    (uint8_t) (0xFF & (liga_set_count >> 8)),
+    (uint8_t) (0xFF & liga_set_count),
+  };
+
+  extend((const char *) liga_set_count_bytes, 2, c);
+}
+
 static unsigned add_pair_pos_2 (unsigned starting_class,
                                 unsigned coverage,
                                 unsigned class_def_1, uint16_t class_def_1_count,
@@ -433,16 +470,13 @@ struct MarkBasePosBuffers
   }
 };
 
-
-
-
-
 static void run_resolve_overflow_test (const char* name,
                                        hb_serialize_context_t& overflowing,
                                        hb_serialize_context_t& expected,
                                        unsigned num_iterations = 0,
                                        bool recalculate_extensions = false,
-                                       hb_tag_t tag = HB_TAG ('G', 'S', 'U', 'B'))
+                                       hb_tag_t tag = HB_TAG ('G', 'S', 'U', 'B'),
+                                       bool check_binary_equivalence = false)
 {
   printf (">>> Testing overflowing resolution for %s\n",
           name);
@@ -452,24 +486,43 @@ static void run_resolve_overflow_test (const char* name,
   graph_t expected_graph (expected.object_graph ());
   if (graph::will_overflow (expected_graph))
   {
+    if (check_binary_equivalence) {
+      printf("when binary equivalence checking is enabled, the expected graph cannot overflow.");
+      hb_always_assert(!check_binary_equivalence);
+    }
     expected_graph.assign_spaces ();
     expected_graph.sort_shortest_distance ();
   }
 
   // Check that overflow resolution succeeds
-  assert (overflowing.offset_overflow ());
-  assert (hb_resolve_graph_overflows (tag,
+  hb_always_assert (overflowing.offset_overflow ());
+  hb_always_assert (hb_resolve_graph_overflows (tag,
                                       num_iterations,
                                       recalculate_extensions,
                                       graph));
 
   // Check the graphs can be serialized.
-  hb_blob_t* out = graph::serialize (graph);
-  assert (out);
-  hb_blob_destroy (out);
-  out = graph::serialize (expected_graph);
-  assert (out);
-  hb_blob_destroy (out);
+  hb_blob_t* out1 = graph::serialize (graph);
+  hb_always_assert (out1);
+  hb_blob_t* out2 = graph::serialize (expected_graph);
+  hb_always_assert (out2);
+  if (check_binary_equivalence) {
+    unsigned l1, l2;
+    const char* d1 = hb_blob_get_data(out1, &l1);
+    const char* d2 = hb_blob_get_data(out2, &l2);
+
+    bool match = (l1 == l2) && (memcmp(d1, d2, l1) == 0);
+    if (!match) {
+      printf("## Result:\n");
+      graph.print();
+      printf("## Expected:\n");
+      expected_graph.print();
+      hb_always_assert(match);
+    }
+  }
+
+  hb_blob_destroy (out1);
+  hb_blob_destroy (out2);
 
   // Check the graphs are equivalent
   graph.normalize ();
@@ -480,7 +533,7 @@ static void run_resolve_overflow_test (const char* name,
     printf("## Result:\n");
     graph.print();
   }
-  assert (graph == expected_graph);
+  hb_always_assert (graph == expected_graph);
 }
 
 static void add_virtual_offset (unsigned id,
@@ -505,6 +558,39 @@ populate_serializer_simple (hb_serialize_context_t* c)
   c->end_serialize();
 }
 
+static void
+populate_serializer_virtual (hb_serialize_context_t* c, bool with_overflow)
+{
+  std::string large_string(50000, 'a');
+  c->start_serialize<char> ();
+
+  unsigned obj_4, obj_5;
+  if (with_overflow) {
+    obj_5 = add_object("55555", 5, c);
+    obj_4 = add_object(large_string.c_str(), 50000, c);
+  } else {
+    obj_4 = add_object(large_string.c_str(), 50000, c);
+    obj_5 = add_object("55555", 5, c);
+  }
+
+  start_object(large_string.c_str(), 20000, c);
+  add_offset(obj_5, c);
+  unsigned obj_3 = c->pop_pack(false);
+
+  start_object("2", 2, c);
+  add_virtual_offset(obj_5, c);
+  unsigned obj_2 = c->pop_pack(false);
+
+  // obj 1
+  start_object("1", 1, c);
+  add_offset(obj_2, c);
+  add_offset(obj_3, c);
+  add_offset(obj_4, c);
+  c->pop_pack(false);
+
+  c->end_serialize();
+}
+
 static void
 populate_serializer_with_overflow (hb_serialize_context_t* c)
 {
@@ -596,6 +682,31 @@ populate_serializer_with_dedup_overflow (hb_serialize_context_t* c)
   c->end_serialize();
 }
 
+static void
+populate_serializer_with_multiple_dedup_overflow (hb_serialize_context_t* c)
+{
+  std::string large_string(70000, 'a');
+  c->start_serialize<char> ();
+
+  unsigned leaf = add_object("def", 3, c);
+
+  constexpr unsigned num_mid_nodes = 20;
+  unsigned mid_nodes[num_mid_nodes];
+  for (unsigned i = 0; i < num_mid_nodes; i++) {
+    start_object(large_string.c_str(), 10000 + i, c);
+    add_offset(leaf, c);
+    mid_nodes[i] = c->pop_pack(false);
+  }
+
+  start_object("abc", 3, c);
+  for (unsigned i = 0; i < num_mid_nodes; i++) {
+    add_wide_offset(mid_nodes[i], c);
+  }
+  c->pop_pack(false);
+
+  c->end_serialize();
+}
+
 static void
 populate_serializer_with_isolation_overflow (hb_serialize_context_t* c)
 {
@@ -754,6 +865,54 @@ populate_serializer_with_isolation_overflow_spaces (hb_serialize_context_t* c)
   c->end_serialize();
 }
 
+static void
+populate_serializer_with_repack_last (hb_serialize_context_t* c, bool with_overflow)
+{
+  std::string large_string(70000, 'c');
+  c->start_serialize<char> ();
+  c->push();
+
+  // Obj E
+  unsigned obj_e_1, obj_e_2;
+  if (with_overflow) {
+    obj_e_1 = add_object("a", 1, c);
+    obj_e_2 = obj_e_1;
+  } else {
+    obj_e_2 = add_object("a", 1, c);
+  }
+
+  // Obj D
+  c->push();
+  add_offset(obj_e_2, c);
+  extend(large_string.c_str(), 30000, c);
+  unsigned obj_d = c->pop_pack(false);
+
+  add_offset(obj_d, c);
+  hb_always_assert(c->last_added_child_index() == obj_d);
+
+  if (!with_overflow) {
+    obj_e_1 = add_object("a", 1, c);
+  }
+
+  // Obj C
+  c->push();
+  add_offset(obj_e_1, c);
+  extend(large_string.c_str(), 40000, c);
+  unsigned obj_c = c->pop_pack(false);
+
+  add_offset(obj_c, c);
+
+  // Obj B
+  unsigned obj_b = add_object("b", 1, c);
+  add_offset(obj_b, c);
+
+  // Obj A
+  c->repack_last(obj_d);
+  c->pop_pack(false);
+
+  c->end_serialize();
+}
+
 static void
 populate_serializer_spaces (hb_serialize_context_t* c, bool with_overflow)
 {
@@ -1479,74 +1638,247 @@ populate_serializer_with_large_mark_base_pos_1 (hb_serialize_context_t* c)
   c->end_serialize();
 }
 
+template<int liga_subst_count,
+         int liga_set_count,
+         int liga_per_set_count,
+         int liga_size>
+static void
+populate_serializer_with_large_liga (hb_serialize_context_t* c, bool sequential_liga_sets)
+{
+  std::string large_string(100000, 'a');
+  c->start_serialize<char> ();
+
+  unsigned liga_subst[liga_subst_count];
+
+  for (unsigned l = 0; l < liga_subst_count; l++) {
+    unsigned coverage_start = 0;
+    unsigned coverage_end = liga_set_count - 1;
+    if (sequential_liga_sets) {
+      coverage_start = l * liga_set_count;
+      coverage_end = (l + 1) * liga_set_count - 1;
+    }
+    unsigned coverage = add_coverage(coverage_start, coverage_end, c);
+
+    unsigned liga[liga_set_count * liga_per_set_count];
+    unsigned liga_set[liga_set_count];
+    for (unsigned i = 0; i < liga_set_count; i++) {
+      for (unsigned j = 0; j < liga_per_set_count; j++)
+      {
+        start_object (large_string.c_str(), liga_size, c);
+        add_virtual_offset(coverage, c);
+        liga[i * liga_per_set_count + j] = c->pop_pack (false);
+      }
+
+      add_liga_set_header(liga_per_set_count, c);
+      add_virtual_offset(coverage, c);
+      for (unsigned j = 0; j < liga_per_set_count; j++)
+      {
+        add_offset(liga[i * liga_per_set_count + j], c);
+      }
+
+      liga_set[i] = c->pop_pack(false);
+    }
+
+    add_liga_header(coverage, liga_set_count, c);
+    for (unsigned i = 0; i < liga_set_count; i++)
+    {
+      add_offset(liga_set[i], c);
+    }
+
+    liga_subst[l] = c->pop_pack(false);
+  }
+
+  for (unsigned l = 0; l < liga_subst_count; l++) {
+    liga_subst[l] = add_extension(liga_subst[l], 4, c);
+  }
+
+  start_lookup (7, liga_subst_count, c);
+  for (unsigned l = 0; l < liga_subst_count; l++) {
+    add_offset(liga_subst[l], c);
+  }
+
+  unsigned lookup = finish_lookup (c);
+
+  unsigned lookup_list = add_lookup_list (&lookup, 1, c);
+
+  add_gsubgpos_header (lookup_list, c);
+
+  c->end_serialize();
+}
+
+static void
+populate_serializer_with_large_liga_overlapping_clone_result (hb_serialize_context_t* c)
+{
+  std::string large_string(100000, 'a');
+  c->start_serialize<char> ();
+
+  constexpr unsigned liga_size = 30000;
+
+  unsigned liga[6];
+  unsigned liga_subst[3];
+  unsigned liga_set[2];
+
+  // LigaSubst 3
+  unsigned coverage = add_coverage(1, 1, c);
+  for (int i = 1; i >= 0; i--) {
+    start_object (large_string.c_str(), liga_size, c);
+    add_virtual_offset(coverage, c);
+    liga[i] = c->pop_pack (false);
+  }
+
+  add_liga_set_header(2, c);
+  add_virtual_offset(coverage, c);
+  for (unsigned i = 0; i < 2; i++)
+    add_offset(liga[i], c);
+  liga_set[0] = c->pop_pack(false);
+
+  add_liga_header(coverage, 1, c);
+  add_offset(liga_set[0], c);
+  liga_subst[2] = c->pop_pack(false);
+
+  // LigaSubst 2
+  coverage = add_coverage(0, 1, c);
+  for (int i = 1; i >= 0; i--) {
+    start_object (large_string.c_str(), liga_size, c);
+    add_virtual_offset(coverage, c);
+    liga[i] = c->pop_pack (false);
+  }
+
+  add_liga_set_header(1, c);
+  add_virtual_offset(coverage, c);
+  add_offset(liga[1], c);
+  liga_set[1] = c->pop_pack(false);
+
+  add_liga_set_header(1, c);
+  add_virtual_offset(coverage, c);
+  add_offset(liga[0], c);
+  liga_set[0] = c->pop_pack(false);
+
+  add_liga_header(coverage, 2, c);
+  add_offset(liga_set[0], c);
+  add_offset(liga_set[1], c);
+  liga_subst[1] = c->pop_pack(false);
+
+  // LigaSubst 1
+  coverage = add_coverage(0, 0, c);
+  for (int i = 1; i >= 0; i--) {
+    start_object (large_string.c_str(), liga_size, c);
+    add_virtual_offset(coverage, c);
+    liga[i] = c->pop_pack (false);
+  }
+
+  add_liga_set_header(2, c);
+  add_virtual_offset(coverage, c);
+  for (unsigned i = 0; i < 2; i++)
+    add_offset(liga[i], c);
+  liga_set[0] = c->pop_pack(false);
+
+  add_liga_header(coverage, 1, c);
+  add_offset(liga_set[0], c);
+  liga_subst[0] = c->pop_pack(false);
+
+  for (int l = 2; l >= 0; l--) {
+    liga_subst[l] = add_extension(liga_subst[l], 4, c);
+  }
+
+  start_lookup (7, 3, c);
+  for (unsigned l = 0; l < 3; l++) {
+    add_offset(liga_subst[l], c);
+  }
+
+  unsigned lookup = finish_lookup (c);
+
+  unsigned lookup_list = add_lookup_list (&lookup, 1, c);
+
+  add_gsubgpos_header (lookup_list, c);
+
+  c->end_serialize();
+}
+
 static void test_sort_shortest ()
 {
   size_t buffer_size = 100;
-  void* buffer = malloc (buffer_size);
-  hb_serialize_context_t c (buffer, buffer_size);
-  populate_serializer_complex_2 (&c);
+  void* buffer_a = malloc (buffer_size);
+  void* buffer_e = malloc (buffer_size);
+  hb_serialize_context_t a (buffer_a, buffer_size);
+  hb_serialize_context_t e (buffer_e, buffer_size);
+  populate_serializer_complex_2 (&a);
 
-  graph_t graph (c.object_graph ());
+  graph_t graph (a.object_graph ());
   graph.sort_shortest_distance ();
-  assert (!graph.in_error ());
+  graph.normalize();
+  hb_always_assert (!graph.in_error ());
 
-  assert(strncmp (graph.object (4).head, "abc", 3) == 0);
-  assert(graph.object (4).real_links.length == 3);
-  assert(graph.object (4).real_links[0].objidx == 2);
-  assert(graph.object (4).real_links[1].objidx == 0);
-  assert(graph.object (4).real_links[2].objidx == 3);
+  // Expected graph
+  e.start_serialize();
+  unsigned jkl = add_object ("jkl", 3, &e);
 
-  assert(strncmp (graph.object (3).head, "mn", 2) == 0);
-  assert(graph.object (3).real_links.length == 0);
+  start_object("ghi", 3, &e);
+  add_offset(jkl, &e);
+  unsigned ghi = e.pop_pack(false);
 
-  assert(strncmp (graph.object (2).head, "def", 3) == 0);
-  assert(graph.object (2).real_links.length == 1);
-  assert(graph.object (2).real_links[0].objidx == 1);
+  start_object("def", 3, &e);
+  add_offset(ghi, &e);
+  unsigned def = e.pop_pack(false);
 
-  assert(strncmp (graph.object (1).head, "ghi", 3) == 0);
-  assert(graph.object (1).real_links.length == 1);
-  assert(graph.object (1).real_links[0].objidx == 0);
+  unsigned mn = add_object("mn", 2, &e);
 
-  assert(strncmp (graph.object (0).head, "jkl", 3) == 0);
-  assert(graph.object (0).real_links.length == 0);
+  start_object ("abc", 3, &e);
+  add_offset (def, &e);
+  add_offset (jkl, &e);
+  add_offset (mn, &e);
+  e.pop_pack (false);
+  e.end_serialize();
 
-  free (buffer);
+  graph_t expected(e.object_graph());
+  expected.normalize();
+
+  assert(expected == graph);
+
+  free (buffer_a);
+  free (buffer_e);
 }
 
 static void test_duplicate_leaf ()
 {
   size_t buffer_size = 100;
-  void* buffer = malloc (buffer_size);
-  hb_serialize_context_t c (buffer, buffer_size);
-  populate_serializer_complex_2 (&c);
+  void* buffer_a = malloc (buffer_size);
+  void* buffer_e = malloc (buffer_size);
+  hb_serialize_context_t a (buffer_a, buffer_size);
+  hb_serialize_context_t e (buffer_e, buffer_size);
+  populate_serializer_complex_2 (&a);
 
-  graph_t graph (c.object_graph ());
+  graph_t graph (a.object_graph ());
   graph.duplicate (4, 1);
+  graph.normalize();
 
-  assert(strncmp (graph.object (5).head, "abc", 3) == 0);
-  assert(graph.object (5).real_links.length == 3);
-  assert(graph.object (5).real_links[0].objidx == 3);
-  assert(graph.object (5).real_links[1].objidx == 4);
-  assert(graph.object (5).real_links[2].objidx == 0);
+  e.start_serialize();
+  unsigned mn = add_object("mn", 2, &e);
+  unsigned jkl_2 = add_object("jkl", 3, &e);
 
-  assert(strncmp (graph.object (4).head, "jkl", 3) == 0);
-  assert(graph.object (4).real_links.length == 0);
+  start_object("ghi", 3, &e);
+  add_offset(jkl_2, &e);
+  unsigned ghi = e.pop_pack(false);
 
-  assert(strncmp (graph.object (3).head, "def", 3) == 0);
-  assert(graph.object (3).real_links.length == 1);
-  assert(graph.object (3).real_links[0].objidx == 2);
+  start_object("def", 3, &e);
+  add_offset(ghi, &e);
+  unsigned def = e.pop_pack(false);
 
-  assert(strncmp (graph.object (2).head, "ghi", 3) == 0);
-  assert(graph.object (2).real_links.length == 1);
-  assert(graph.object (2).real_links[0].objidx == 1);
+  unsigned jkl_1 = add_object("jkl", 3, &e);
 
-  assert(strncmp (graph.object (1).head, "jkl", 3) == 0);
-  assert(graph.object (1).real_links.length == 0);
+  start_object("abc", 3, &e);
+  add_offset(def, &e);
+  add_offset(jkl_1, &e);
+  add_offset(mn, &e);
+  e.pop_pack(false);
 
-  assert(strncmp (graph.object (0).head, "mn", 2) == 0);
-  assert(graph.object (0).real_links.length == 0);
+  graph_t expected(e.object_graph());
+  expected.normalize();
 
-  free (buffer);
+  assert(expected == graph);
+
+  free (buffer_a);
+  free (buffer_e);
 }
 
 static void test_duplicate_interior ()
@@ -1559,33 +1891,33 @@ static void test_duplicate_interior ()
   graph_t graph (c.object_graph ());
   graph.duplicate (3, 2);
 
-  assert(strncmp (graph.object (6).head, "abc", 3) == 0);
-  assert(graph.object (6).real_links.length == 3);
-  assert(graph.object (6).real_links[0].objidx == 4);
-  assert(graph.object (6).real_links[1].objidx == 2);
-  assert(graph.object (6).real_links[2].objidx == 1);
+  hb_always_assert(strncmp (graph.object (6).head, "jkl", 3) == 0);
+  hb_always_assert(graph.object (6).real_links.length == 1);
+  hb_always_assert(graph.object (6).real_links[0].objidx == 0);
 
-  assert(strncmp (graph.object (5).head, "jkl", 3) == 0);
-  assert(graph.object (5).real_links.length == 1);
-  assert(graph.object (5).real_links[0].objidx == 0);
+  hb_always_assert(strncmp (graph.object (5).head, "abc", 3) == 0);
+  hb_always_assert(graph.object (5).real_links.length == 3);
+  hb_always_assert(graph.object (5).real_links[0].objidx == 4);
+  hb_always_assert(graph.object (5).real_links[1].objidx == 2);
+  hb_always_assert(graph.object (5).real_links[2].objidx == 1);
 
-  assert(strncmp (graph.object (4).head, "def", 3) == 0);
-  assert(graph.object (4).real_links.length == 1);
-  assert(graph.object (4).real_links[0].objidx == 3);
+  hb_always_assert(strncmp (graph.object (4).head, "def", 3) == 0);
+  hb_always_assert(graph.object (4).real_links.length == 1);
+  hb_always_assert(graph.object (4).real_links[0].objidx == 3);
 
-  assert(strncmp (graph.object (3).head, "ghi", 3) == 0);
-  assert(graph.object (3).real_links.length == 1);
-  assert(graph.object (3).real_links[0].objidx == 5);
+  hb_always_assert(strncmp (graph.object (3).head, "ghi", 3) == 0);
+  hb_always_assert(graph.object (3).real_links.length == 1);
+  hb_always_assert(graph.object (3).real_links[0].objidx == 6);
 
-  assert(strncmp (graph.object (2).head, "jkl", 3) == 0);
-  assert(graph.object (2).real_links.length == 1);
-  assert(graph.object (2).real_links[0].objidx == 0);
+  hb_always_assert(strncmp (graph.object (2).head, "jkl", 3) == 0);
+  hb_always_assert(graph.object (2).real_links.length == 1);
+  hb_always_assert(graph.object (2).real_links[0].objidx == 0);
 
-  assert(strncmp (graph.object (1).head, "mn", 2) == 0);
-  assert(graph.object (1).real_links.length == 0);
+  hb_always_assert(strncmp (graph.object (1).head, "mn", 2) == 0);
+  hb_always_assert(graph.object (1).real_links.length == 0);
 
-  assert(strncmp (graph.object (0).head, "opqrst", 6) == 0);
-  assert(graph.object (0).real_links.length == 0);
+  hb_always_assert(strncmp (graph.object (0).head, "opqrst", 6) == 0);
+  hb_always_assert(graph.object (0).real_links.length == 0);
 
   free (buffer);
 }
@@ -1604,7 +1936,7 @@ test_serialize ()
   free (buffer_1);
 
   hb_bytes_t actual = out->as_bytes ();
-  assert (actual == expected);
+  hb_always_assert (actual == expected);
   expected.fini ();
   hb_blob_destroy (out);
 }
@@ -1614,10 +1946,11 @@ static void test_will_overflow_1 ()
   size_t buffer_size = 100;
   void* buffer = malloc (buffer_size);
   hb_serialize_context_t c (buffer, buffer_size);
+
   populate_serializer_complex_2 (&c);
   graph_t graph (c.object_graph ());
 
-  assert (!graph::will_overflow (graph, nullptr));
+  hb_always_assert (!graph::will_overflow (graph, nullptr));
 
   free (buffer);
 }
@@ -1630,7 +1963,7 @@ static void test_will_overflow_2 ()
   populate_serializer_with_overflow (&c);
   graph_t graph (c.object_graph ());
 
-  assert (graph::will_overflow (graph, nullptr));
+  hb_always_assert (graph::will_overflow (graph, nullptr));
 
   free (buffer);
 }
@@ -1643,7 +1976,7 @@ static void test_will_overflow_3 ()
   populate_serializer_with_dedup_overflow (&c);
   graph_t graph (c.object_graph ());
 
-  assert (graph::will_overflow (graph, nullptr));
+  hb_always_assert (graph::will_overflow (graph, nullptr));
 
   free (buffer);
 }
@@ -1657,9 +1990,9 @@ static void test_resolve_overflows_via_sort ()
   graph_t graph (c.object_graph ());
 
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG_NONE);
-  assert (out);
+  hb_always_assert (out);
   hb_bytes_t result = out->as_bytes ();
-  assert (result.length == (80000 + 3 + 3 * 2));
+  hb_always_assert (result.length == (80000 + 3 + 3 * 2));
 
   free (buffer);
   hb_blob_destroy (out);
@@ -1674,9 +2007,24 @@ static void test_resolve_overflows_via_duplication ()
   graph_t graph (c.object_graph ());
 
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG_NONE);
-  assert (out);
+  hb_always_assert (out);
   hb_bytes_t result = out->as_bytes ();
-  assert (result.length == (10000 + 2 * 2 + 60000 + 2 + 3 * 2));
+  hb_always_assert (result.length == (10000 + 2 * 2 + 60000 + 2 + 3 * 2));
+
+  free (buffer);
+  hb_blob_destroy (out);
+}
+
+static void test_resolve_overflows_via_multiple_duplications ()
+{
+  size_t buffer_size = 300000;
+  void* buffer = malloc (buffer_size);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_with_multiple_dedup_overflow (&c);
+  graph_t graph (c.object_graph ());
+
+  hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG_NONE, 5);
+  hb_always_assert (out);
 
   free (buffer);
   hb_blob_destroy (out);
@@ -1709,11 +2057,11 @@ static void test_resolve_overflows_via_isolation ()
   populate_serializer_with_isolation_overflow (&c);
   graph_t graph (c.object_graph ());
 
-  assert (c.offset_overflow ());
+  hb_always_assert (c.offset_overflow ());
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG ('G', 'S', 'U', 'B'), 0);
-  assert (out);
+  hb_always_assert (out);
   hb_bytes_t result = out->as_bytes ();
-  assert (result.length == (1 + 10000 + 60000 + 1 + 1
+  hb_always_assert (result.length == (1 + 10000 + 60000 + 1 + 1
                             + 4 + 3 * 2));
 
   free (buffer);
@@ -1784,14 +2132,14 @@ static void test_resolve_overflows_via_isolation_spaces ()
   populate_serializer_with_isolation_overflow_spaces (&c);
   graph_t graph (c.object_graph ());
 
-  assert (c.offset_overflow ());
+  hb_always_assert (c.offset_overflow ());
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG ('G', 'S', 'U', 'B'), 0);
-  assert (out);
+  hb_always_assert (out);
   hb_bytes_t result = out->as_bytes ();
 
   unsigned expected_length = 3 + 2 * 60000; // objects
   expected_length += 2 * 4 + 2 * 2; // links
-  assert (result.length == expected_length);
+  hb_always_assert (result.length == expected_length);
 
   free (buffer);
   hb_blob_destroy (out);
@@ -1805,9 +2153,9 @@ static void test_resolve_mixed_overflows_via_isolation_spaces ()
   populate_serializer_with_24_and_32_bit_offsets (&c);
   graph_t graph (c.object_graph ());
 
-  assert (c.offset_overflow ());
+  hb_always_assert (c.offset_overflow ());
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG ('G', 'S', 'U', 'B'), 0);
-  assert (out);
+  hb_always_assert (out);
   hb_bytes_t result = out->as_bytes ();
 
   unsigned expected_length =
@@ -1821,7 +2169,7 @@ static void test_resolve_mixed_overflows_via_isolation_spaces ()
       4 * 3 +  // 24
       4 * 2;   // 16
 
-  assert (result.length == expected_length);
+  hb_always_assert (result.length == expected_length);
 
   free (buffer);
   hb_blob_destroy (out);
@@ -1831,12 +2179,12 @@ static void test_resolve_with_extension_promotion ()
 {
   size_t buffer_size = 200000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_extension_promotion (&c);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_extension_promotion (&e, 3);
 
@@ -1853,12 +2201,12 @@ static void test_resolve_with_shared_extension_promotion ()
 {
   size_t buffer_size = 200000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_extension_promotion (&c, 0, true);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_extension_promotion (&e, 3, true);
 
@@ -1875,12 +2223,12 @@ static void test_resolve_with_basic_pair_pos_1_split ()
 {
   size_t buffer_size = 200000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_large_pair_pos_1 <1, 4>(&c);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_large_pair_pos_1 <2, 2>(&e, true);
 
@@ -1898,12 +2246,12 @@ static void test_resolve_with_extension_pair_pos_1_split ()
 {
   size_t buffer_size = 200000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_large_pair_pos_1 <1, 4>(&c, true);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_large_pair_pos_1 <2, 2>(&e, true);
 
@@ -1921,12 +2269,12 @@ static void test_resolve_with_basic_pair_pos_2_split ()
 {
   size_t buffer_size = 300000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_large_pair_pos_2 <1, 4, 3000>(&c);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_large_pair_pos_2 <2, 2, 3000>(&e, true);
 
@@ -1944,14 +2292,14 @@ static void test_resolve_with_close_to_limit_pair_pos_2_split ()
 {
   size_t buffer_size = 300000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
-  populate_serializer_with_large_pair_pos_2 <1, 1596, 10>(&c, true, false, false);
+  populate_serializer_with_large_pair_pos_2 <1, 1636, 10>(&c, true, false, false);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
-  populate_serializer_with_large_pair_pos_2 <2, 798, 10>(&e, true, false, false);
+  populate_serializer_with_large_pair_pos_2 <2, 818, 10>(&e, true, false, false);
 
   run_resolve_overflow_test ("test_resolve_with_close_to_limit_pair_pos_2_split",
                              c,
@@ -1967,12 +2315,12 @@ static void test_resolve_with_pair_pos_2_split_with_device_tables ()
 {
   size_t buffer_size = 300000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_large_pair_pos_2 <1, 4, 2000>(&c, false, true);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_large_pair_pos_2 <2, 2, 2000>(&e, true, true);
 
@@ -1990,12 +2338,12 @@ static void test_resolve_with_basic_mark_base_pos_1_split ()
 {
   size_t buffer_size = 200000;
   void* buffer = malloc (buffer_size);
-  assert (buffer);
+  hb_always_assert (buffer);
   hb_serialize_context_t c (buffer, buffer_size);
   populate_serializer_with_large_mark_base_pos_1 <40, 10, 110, 1>(&c);
 
   void* expected_buffer = malloc (buffer_size);
-  assert (expected_buffer);
+  hb_always_assert (expected_buffer);
   hb_serialize_context_t e (expected_buffer, buffer_size);
   populate_serializer_with_large_mark_base_pos_1 <40, 10, 110, 2>(&e);
 
@@ -2009,6 +2357,75 @@ static void test_resolve_with_basic_mark_base_pos_1_split ()
   free (expected_buffer);
 }
 
+static void test_resolve_with_basic_liga_split ()
+{
+  size_t buffer_size = 200000;
+  void* buffer = malloc (buffer_size);
+  hb_always_assert (buffer);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_with_large_liga<1, 1, 2, 40000>(&c, false);
+
+  void* expected_buffer = malloc (buffer_size);
+  hb_always_assert (expected_buffer);
+  hb_serialize_context_t e (expected_buffer, buffer_size);
+  populate_serializer_with_large_liga<2, 1, 1, 40000>(&e, false);
+
+  run_resolve_overflow_test ("test_resolve_with_basic_liga_split",
+                             c,
+                             e,
+                             20,
+                             true,
+                             HB_TAG('G', 'S', 'U', 'B'));
+  free (buffer);
+  free (expected_buffer);
+}
+
+static void test_resolve_with_liga_split_move ()
+{
+  size_t buffer_size = 400000;
+  void* buffer = malloc (buffer_size);
+  hb_always_assert (buffer);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_with_large_liga<1, 6, 2, 16000>(&c, true);
+
+  void* expected_buffer = malloc (buffer_size);
+  hb_always_assert (expected_buffer);
+  hb_serialize_context_t e (expected_buffer, buffer_size);
+  populate_serializer_with_large_liga<3, 2, 2, 16000>(&e, true);
+
+  run_resolve_overflow_test ("test_resolve_with_liga_split_move",
+                             c,
+                             e,
+                             20,
+                             true,
+                             HB_TAG('G', 'S', 'U', 'B'));
+  free (buffer);
+  free (expected_buffer);
+}
+
+static void test_resolve_with_liga_split_overlapping_clone ()
+{
+  size_t buffer_size = 400000;
+  void* buffer = malloc (buffer_size);
+  hb_always_assert (buffer);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_with_large_liga<1, 2, 3, 30000>(&c, true);
+
+  void* expected_buffer = malloc (buffer_size);
+  hb_always_assert (expected_buffer);
+  hb_serialize_context_t e (expected_buffer, buffer_size);
+  populate_serializer_with_large_liga_overlapping_clone_result(&e);
+
+  run_resolve_overflow_test ("test_resolve_with_liga_split_overlapping_clone",
+                             c,
+                             e,
+                             20,
+                             true,
+                             HB_TAG('G', 'S', 'U', 'B'));
+  free (buffer);
+  free (expected_buffer);
+}
+
 static void test_resolve_overflows_via_splitting_spaces ()
 {
   size_t buffer_size = 160000;
@@ -2077,15 +2494,15 @@ static void test_virtual_link ()
   populate_serializer_virtual_link (&c);
 
   hb_blob_t* out = hb_resolve_overflows (c.object_graph (), HB_TAG_NONE);
-  assert (out);
+  hb_always_assert (out);
 
   hb_bytes_t result = out->as_bytes ();
-  assert (result.length == 5 + 4 * 2);
-  assert (result[0]  == 'a');
-  assert (result[5]  == 'c');
-  assert (result[8]  == 'e');
-  assert (result[9]  == 'b');
-  assert (result[12] == 'd');
+  hb_always_assert (result.length == 5 + 4 * 2);
+  hb_always_assert (result[0]  == 'a');
+  hb_always_assert (result[5]  == 'c');
+  hb_always_assert (result[8]  == 'e');
+  hb_always_assert (result[9]  == 'b');
+  hb_always_assert (result[12] == 'd');
 
   free (buffer);
   hb_blob_destroy (out);
@@ -2111,7 +2528,7 @@ test_shared_node_with_virtual_links ()
   add_virtual_offset (obj_c, &c);
   unsigned obj_d_2 = c.pop_pack ();
 
-  assert (obj_d_1 == obj_d_2);
+  hb_always_assert (obj_d_1 == obj_d_2);
 
   start_object ("a", 1, &c);
   add_offset (obj_b, &c);
@@ -2121,12 +2538,63 @@ test_shared_node_with_virtual_links ()
   c.pop_pack ();
   c.end_serialize ();
 
-  assert(c.object_graph() [obj_d_1]->virtual_links.length == 2);
-  assert(c.object_graph() [obj_d_1]->virtual_links[0].objidx == obj_b);
-  assert(c.object_graph() [obj_d_1]->virtual_links[1].objidx == obj_c);
+  hb_always_assert(c.object_graph() [obj_d_1]->virtual_links.length == 2);
+  hb_always_assert(c.object_graph() [obj_d_1]->virtual_links[0].objidx == obj_b);
+  hb_always_assert(c.object_graph() [obj_d_1]->virtual_links[1].objidx == obj_c);
   free(buffer);
 }
 
+static void
+test_repack_last ()
+{
+  size_t buffer_size = 200000;
+  void* buffer = malloc (buffer_size);
+  hb_always_assert (buffer);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_with_repack_last (&c, true);
+
+  void* expected_buffer = malloc (buffer_size);
+  hb_always_assert (expected_buffer);
+  hb_serialize_context_t e (expected_buffer, buffer_size);
+  populate_serializer_with_repack_last (&e, false);
+
+  run_resolve_overflow_test ("test_repack_last",
+                             c,
+                             e,
+                             20,
+                             false,
+                             HB_TAG('a', 'b', 'c', 'd'),
+                             true);
+
+  free (buffer);
+  free (expected_buffer);
+}
+
+static void
+test_dont_duplicate_virtual ()
+{
+  size_t buffer_size = 200000;
+  void* buffer = malloc (buffer_size);
+  hb_always_assert (buffer);
+  hb_serialize_context_t c (buffer, buffer_size);
+  populate_serializer_virtual (&c, true);
+
+  void* expected_buffer = malloc (buffer_size);
+  hb_always_assert (expected_buffer);
+  hb_serialize_context_t e (expected_buffer, buffer_size);
+  populate_serializer_virtual (&e, false);
+
+  run_resolve_overflow_test ("test_dont_duplicate_virtual",
+                             c,
+                             e,
+                             20,
+                             false,
+                             HB_TAG('a', 'b', 'c', 'd'),
+                             true);
+
+  free (buffer);
+  free (expected_buffer);
+}
 
 // TODO(garretrieger): update will_overflow tests to check the overflows array.
 // TODO(garretrieger): add tests for priority raising.
@@ -2141,6 +2609,7 @@ main (int argc, char **argv)
   test_will_overflow_3 ();
   test_resolve_overflows_via_sort ();
   test_resolve_overflows_via_duplication ();
+  test_resolve_overflows_via_multiple_duplications ();
   test_resolve_overflows_via_priority ();
   test_resolve_overflows_via_space_assignment ();
   test_resolve_overflows_via_isolation ();
@@ -2154,6 +2623,7 @@ main (int argc, char **argv)
   test_duplicate_leaf ();
   test_duplicate_interior ();
   test_virtual_link ();
+  test_repack_last();
   test_shared_node_with_virtual_links ();
   test_resolve_with_extension_promotion ();
   test_resolve_with_shared_extension_promotion ();
@@ -2163,6 +2633,10 @@ main (int argc, char **argv)
   test_resolve_with_pair_pos_2_split_with_device_tables ();
   test_resolve_with_close_to_limit_pair_pos_2_split ();
   test_resolve_with_basic_mark_base_pos_1_split ();
+  test_resolve_with_basic_liga_split ();
+  test_resolve_with_liga_split_move ();
+  test_dont_duplicate_virtual ();
+  test_resolve_with_liga_split_overlapping_clone ();
 
   // TODO(grieger): have run overflow tests compare graph equality not final packed binary.
   // TODO(grieger): split test where multiple subtables in one lookup are split to test link ordering.
